from __future__ import annotations
import asyncio
from collections import defaultdict
from typing import Callable, List, Dict
from ..util import exceptionToStr

_emitter: EventEmitter = None


def getEmitter(onError: Callable = None) -> EventEmitter:
    global _emitter
    if _emitter is None:
        _emitter = EventEmitter(onError)
        if not onError:
            raise Exception(
                "getEmitter() must be called with onError callable as an argument on first call!"
            )
    return _emitter


class EventEmitter:
    def __init__(self, onError: Callable):
        if _emitter is not None:
            raise Exception(
                "EventEmitter must be a singleton! Use getEmitter() to create/get the instance."
            )
        self._listeners: Dict[str, List[Callable]] = defaultdict(list)
        self.onError: Callable[[str], None] = onError

    def on(self, event, listener):
        """Subscribe a listener to an event."""
        if not callable(listener):
            raise TypeError("Listener must be a callable (sync or async function)")
        self._listeners[event].append(listener)

    def off(self, event, listener):
        """Unsubscribe a listener from an event."""
        if listener in self._listeners[event]:
            self._listeners[event].remove(listener)

    def emit(self, event, *args, **kwargs):
        """Emit an event and handle all listeners (sync or async)."""
        if event not in self._listeners:
            return

        errorStrs = []

        for listener in self._listeners[event]:
            if asyncio.iscoroutinefunction(listener):
                # Async listener
                task = asyncio.create_task(listener(event, *args, **kwargs))
                task.add_done_callback(
                    lambda task: self._handle_task_exception(task, event)
                )
            else:
                # Sync listener - execute it directly
                try:
                    listener(event, *args, **kwargs)
                except Exception as e:
                    errorStrs.append(exceptionToStr(e))
        if errorStrs:
            self.onError(event, *errorStrs)

    def _handle_task_exception(self, task: asyncio.Task, event: str):
        """Handle task exceptions."""
        try:
            task.result()  # This will raise the exception if the task failed
        except Exception as e:
            self.onError(event, exceptionToStr(e))



from enum import Enum


class LogLevel(Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    CRITICAL = "CRITICAL"
    WARNING = "WARNING"


class Color(Enum):
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    PURPLE = "\033[1m\033[34m"
    PINK = "\033[35m"
    CYAN = "\033[1m\033[96m"
    WHITE = "\033[37m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    ENDC = "\033[0m"


MAX_VALUE_LENGTH_IN_TABLES = 50



from ..loggingConsts import LogLevel
from ...util import ensureDirExists


class Logger:
    def __init__(
        self,
        logDirectory: str,
        fileName: str,
        logLevel: LogLevel,
        printToConsole: bool = True,
    ):
        # Convert logLevel to enum if it's a string
        if isinstance(logLevel, str):
            try:
                logLevel = LogLevel[logLevel]
            except KeyError:
                raise ValueError(
                    f"Invalid log level: {logLevel}. Expected one of {list(LogLevel.__members__.keys())}"
                )

        elif not isinstance(logLevel, LogLevel):
            raise TypeError(
                f"logLevel must be an instance of LogLevel or a valid string, got {type(logLevel).__name__}"
            )

        ensureDirExists(logDirectory)
        self.filePath = f"{logDirectory}/{fileName}"
        self.logLevel: LogLevel = logLevel
        self.printToConsole: bool = printToConsole

    def clearLog(self):
        with open(self.filePath, "w") as file:
            pass



from __future__ import annotations

from typing import Optional, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from jgmd.logging import Color
    from .consts import LogType


class LogEntry:
    def __init__(
        self,
        lambdaToLog: Callable[[], str],
        logType: "LogType",
        color: Optional["Color"] = None,
    ):
        self.lambdaToLog = lambdaToLog
        self.logType = logType
        self.color = color



from typing import Callable
from ..loggingConsts import LogLevel, Color
from .logger import Logger


class FreeTextLogger(Logger):
    def __init__(
        self,
        logDirectory: str,
        fileName: str,
        logLevel: LogLevel,
        printToConsole: bool = True,
    ):
        super().__init__(logDirectory, fileName, logLevel, printToConsole)
        self.printToConsole: bool = printToConsole

    def _log(self, s: str, color: Color = ""):
        if self.printToConsole:
            print(f"{color}{s}{Color.ENDC.value}")
        with open(self.filePath, "a") as file:
            file.write(s + "\n")

    def logDebug(self, getStrToLog: Callable[[], str], color: Color = None):
        if self.logLevel == LogLevel.DEBUG:
            s = getStrToLog()
            color = color or Color.PINK
            self._log(s, color.value)

    def logInfo(self, getStrToLog: Callable[[], str], color: Color = None):
        if self.logLevel == LogLevel.DEBUG or self.logLevel == LogLevel.INFO:
            s = getStrToLog()
            color = color or Color.CYAN
            self._log(s, color.value)

    def logWarning(self, getStrToLog: Callable[[], str], color: Color = None):
        if (
            self.logLevel == LogLevel.DEBUG
            or self.logLevel == LogLevel.INFO
            or self.logLevel == LogLevel.WARNING
        ):
            s = getStrToLog()
            color = color or Color.YELLOW
            self._log(s, color.value)

    def logError(self, getStrToLog: Callable[[], str], color: Color = None):
        s = getStrToLog()
        color = color or Color.RED
        self._log(s, color.value)

    def logSuccessful(self, getStrToLog: Callable[[], str], color: Color = None):
        s = getStrToLog()
        color = color or Color.GREEN
        self._log(s, color.value)



from __future__ import annotations

from typing import Optional, List, Callable, TYPE_CHECKING
from .logEntry import LogEntry
from .consts import LogType

if TYPE_CHECKING:
    from jgmd.logging import FreeTextLogger
    from jgmd.logging import Color


class LazyLogBuffer:
    def __init__(self):
        self.entries: List[LogEntry] = []

    def lazyLogDebug(
        self, lambdaToLog: Callable[[], str], color: Optional["Color"] = None
    ):
        self.entries.append(LogEntry(lambdaToLog, LogType.DEBUG, color))

    def lazyLogInfo(
        self, lambdaToLog: Callable[[], str], color: Optional["Color"] = None
    ):
        self.entries.append(LogEntry(lambdaToLog, LogType.INFO, color))

    def lazyLogWarning(
        self, lambdaToLog: Callable[[], str], color: Optional["Color"] = None
    ):
        self.entries.append(LogEntry(lambdaToLog, LogType.WARNING, color))

    def lazyLogError(
        self, lambdaToLog: Callable[[], str], color: Optional["Color"] = None
    ):
        self.entries.append(LogEntry(lambdaToLog, LogType.ERROR, color))

    def lazyLogSuccessful(
        self, lambdaToLog: Callable[[], str], color: Optional["Color"] = None
    ):
        self.entries.append(LogEntry(lambdaToLog, LogType.SUCCESS, color))

    def logAll(self, logger: "FreeTextLogger"):
        for entry in self.entries:
            if entry.logType == LogType.DEBUG:
                logger.logDebug(entry.lambdaToLog, entry.color)
            elif entry.logType == LogType.INFO:
                logger.logInfo(entry.lambdaToLog, entry.color)
            elif entry.logType == LogType.WARNING:
                logger.logWarning(entry.lambdaToLog, entry.color)
            elif entry.logType == LogType.ERROR:
                logger.logError(entry.lambdaToLog, entry.color)
            elif entry.logType == LogType.SUCCESS:
                logger.logSuccessful(entry.lambdaToLog, entry.color)



from ..loggingUtil import makeTable

# from collections.abc import Iterable
from .logger import Logger
from ..loggingConsts import LogLevel


class TableLogger(Logger):

    def __init__(
        self,
        logDirectory: str,
        fileName: str,
        logLevel: LogLevel,
        printToConsole: bool = True,
    ):
        super().__init__(logDirectory, fileName, logLevel, printToConsole)

    def getHeaders(self):
        raise NotImplementedError("getHeaders() must be implemented by subclass")

    def getRowData(self, data):
        raise NotImplementedError("getRowData() must be implemented by subclass")

    def clearLog(self):
        super().clearLog()
        self.writeHeaders()

    def writeHeaders(self):
        with open(self.filePath, "a") as file:
            headerStr = ",".join(self.getHeaders())
            file.write(headerStr + "\n")

    def logRowToFile(self, data):
        with open(self.filePath, "a") as file:
            data = self.getRowData(data)
            rowStr = ",".join([str(item) for item in data])
            file.write(rowStr + "\n")

    def getTableStr(self, datas: list, label: str = None):
        if len(datas) == 0:
            return ""
        # if not isinstance(datas[0], Iterable):
        #     datas = [[data] for data in datas]

        data = [self.getRowData(data) for data in datas]

        headers = self.getHeaders()
        return makeTable(data, headers, label=label)



import os
from dotenv import load_dotenv
from pydantic import BaseModel
from pathlib import Path
from typing import Type, TypeVar, Optional

T = TypeVar("T", bound=BaseModel)


def loadEnv(pydanticModel: Type[T], dotenvPath: Optional[str] = None) -> T:
    """
    Load environment variables into a given Pydantic model.

    Args:
        pydanticModel (Type[T]): A Pydantic model class.
        dotenvPath (Optional[str]): Optional path to a .env file.

    Returns:
        T: An instance of the model populated with environment variables.
    """
    load_dotenv(dotenv_path=dotenvPath, override=True)

    env_values = {
        field_name: os.getenv(field_name)
        for field_name in pydanticModel.model_fields.keys()
    }

    return pydanticModel(**env_values)


def loadLocalEnv(pydanticModel: Type[T], filePath: str) -> T:
    """
    Load environment variables from a .env file co-located with the filePath.
    e.g. to load a .env colocated with a given script/module, pass filePath=__file__

    Args:
        pydanticModel (Type[T]): A Pydantic model class.

    Returns:
        T: An instance of the model populated with environment variables.
    """
    dotenvPath = Path(filePath).parent / ".env"
    return loadEnv(pydanticModel, dotenvPath=dotenvPath)



from typing import List, Callable

from ..logging import FreeTextLogger, LogLevel
from .errorUtil import getErrorHandlingDecorators

class AppLogger:
    debugLogger: FreeTextLogger
    errorLogger: FreeTextLogger

    @staticmethod
    def init(logLevel:LogLevel=LogLevel.CRITICAL, logDirectory:str="./logs", debugLogFileName:str="debug.log", errorLogFileName="error.log",printToConsole=True) -> None:
        AppLogger.debugLogger = FreeTextLogger(
            logDirectory=logDirectory,
            fileName=debugLogFileName,
            logLevel=logLevel,
            printToConsole=printToConsole,
        )
        AppLogger.errorLogger = FreeTextLogger(
            logDirectory=logDirectory,
            fileName=errorLogFileName,
            logLevel=LogLevel[logLevel],
            printToConsole=printToConsole,
        )
    @staticmethod
    def clearLogs():
        AppLogger.debugLogger.clearLog()
        AppLogger.errorLogger.clearLog()


class AppErrors:
    errors: List[str] = []
    handleError: Callable[[Callable], Callable]
    handleErrorAsync: Callable[[Callable], Callable]

    @staticmethod
    def init() -> None:
        handleError, handleErrorAsync = getErrorHandlingDecorators(AppErrors.onAppError)
        AppErrors.handleError = handleError
        AppErrors.handleErrorAsync = handleErrorAsync

    @staticmethod
    def onAppError(errorStr: str):
        AppLogger.errorLogger.logError(lambda: errorStr)
        AppErrors.errors.append(errorStr)

    @staticmethod
    def onEventError(event: str, errorStr: str):
        AppLogger.errorLogger.logError(
            lambda: f"Error during event={event}:\n{errorStr}"
        )
        AppErrors.errors.append(errorStr)

    @staticmethod
    def hasErrors():
        return len(AppErrors.errors) > 0